/*  Greenspan Discrete Mechanics for a system of massive particles.	*/
/*  This records the system state and provides methods for initializing	*/
/*  updating same.							*/

import java.lang.Math;
import java.io.*;
import java.net.*;

public class Grnspan
 {
  public int NB,		/* Number of bodies allowed	*/
	     Nc,		/* Number currently defined	*/
	  limit,		/* Max iterations/time step	*/
	   fail;		/* No. of convergence failures	*/

  /* State data, [NB][3] or [NB] : */

  public double X0[][],		/* Current positions		*/
		X1[][],		/* Next positions		*/
		V0[][],		/* Current velocities		*/
		V1[][],		/* Next velocities		*/
		Ac[][],		/* Accelerations		*/
	        Ms[],		/* Masses			*/
		Mg[],		/* Masses * G			*/
		G,		/* Constant of gravitation	*/
		stime,		/* System time for X0, V0	*/
		deltat,		/* Current delta-t		*/
		epsilon;	/* Convergence criterion	*/


/************************************************************************/

Grnspan(int nb, double t0)	/* Number of bodies, starting time */
 { NB = nb;  stime = t0;  Nc = 0;
   X0 = new double[NB][3];  X1 = new double[NB][3];
   V0 = new double[NB][3];  V1 = new double[NB][3];
   Ac = new double[NB][3];
   Ms = new double[NB];      Mg = new double[NB];
 }


/************************************************************************/
/*  Load (more) bodies from file.  Leaves masses in original units	*/
/*  in Ms, masses * G (from file) in Mg.  Current version is for	*/
/*  application with file in accessible directory.			*/

void loadBodies(String file)
 { try
    { StreamTokenizer data =
	   new StreamTokenizer(new BufferedInputStream(
		   new FileInputStream(file)));
		/* new URL( getCodeBase(), file).openStream())); */
      data.parseNumbers();  data.slashSlashComments(true);
      int tok = data.nextToken();  int fn = (int) data.nval;
      if ( Nc + fn > NB )	/* Too many bodies in file */
       System.out.println("Too many bodies in " + file + " - " + fn);
      else	/* Read positions, then velocities, then masses, then G */
       { int fx = Nc + fn;  int i, u;
	 for ( i = Nc;  i < fx;  i++ )
	  for ( u = 0;  u < 3;  u++ )
	   { tok = data.nextToken();  X0[i][u] =  data.nval; }
 	 for ( i = Nc;  i < fx;  i++ )
	  for ( u = 0;  u < 3;  u++ )
	   { tok = data.nextToken();  V0[i][u] =  data.nval; }
 	 for ( i = Nc;  i < fx;  i++ )
	  { tok = data.nextToken();  Ms[i] =  data.nval; }
	 tok = data.nextToken();  G = data.nval;
	 /* Scale masses */
	 for ( i = Nc;  i < fx;  i++ )  Mg[i] = G*Ms[i];
	 Nc = fx;			/* Record new particles */
       }
   }
  catch ( IOException ex)
   { System.out.println("Couldn't read file for bodies - "
			  + ex.getMessage());
   }
 }


/************************************************************************/
/*  Install Body b as next particle					*/

public void installBody(Body b)
  {  if ( Nc < NB )
      { X0[Nc][0] = b.X[0];  X0[Nc][1] = b.X[1];  X0[Nc][2] = b.X[2];
        V0[Nc][0] = b.V[0];  V0[Nc][1] = b.V[1];  V0[Nc][2] = b.V[2];
	Ms[Nc] = b.mass;  Mg[Nc] = Ms[Nc]*G;
        Nc += 1;
      }
  }


/************************************************************************/
/*  Set Ac = 0.0 (all components)					*/

void clearAc()
 { for ( int i = 0;  i < Nc;  i++ )		/* Ac = 0.0 */
    { Ac[i][0] = 0.0;  Ac[i][1] = 0.0;  Ac[i][2] = 0.0; }
 }


/************************************************************************/
/*  Calculate Greenspan acceleration for current X0, X1, leaving result	*/
/*  in Ac.								*/

void greenAccel()
 { double Dij[],	/* (X0[j] - X0[i]) + (X1[j] - X1[i])	*/
	  r0,		/* || X0[j] - X0[i] ||			*/
	  r1,		/* || X1[j] - X1[i] ||			*/
	  fi,		/* Multiplier for particle i		*/
	  fj,		/* Multiplier for particle j		*/
	  w, v;		/* Miscellaneous			*/
   int i, j, u;		/* Loop counters			*/

   Dij = new double[3];  clearAc();
   /* Loop over pairs of particles */
   for ( i = 0;  i < Nc-1;  i++ )
    for ( j = i+1;  j < Nc;  j++ )
     { r0 = 0.0;  r1 = 0.0;		/* To sum squares of diplacements */
       for ( u = 0;  u < 3;  u++ )
        { w = X0[j][u] - X0[i][u];  r0 += w*w;
	  v = X1[j][u] - X1[i][u];  r1 += v*v;
	  Dij[u] = w + v;
	}
       r0 = Math.sqrt(r0);  r1 = Math.sqrt(r1);  w = r0*r1*(r0 + r1);
       fi = Mg[j]/w;   fj = Mg[i]/w;
       for ( u = 0;  u < 3;  u++ )
	{ Ac[i][u] += fi*Dij[u];  Ac[j][u] -= fj*Dij[u]; }
     }
  }


/************************************************************************/
/*  Update X1, V1 using current Ac and deltat, returns max absolute	*/
/*  change in any component.						*/

double updateXV()
 { double mx, w, hdt = 0.5*deltat;  int i, u;

   mx = 0.0;					/* Maximum change so far  */
   for ( i = 0;  i < Nc;  i++ ) 		/* Loop over particles    */
    for ( u = 0;  u < 3;  u++ )
     { w  = V0[i][u] + deltat * Ac[i][u];	/* New velocity component */
       mx = Math.max(mx, Math.abs(V1[i][u] - w));	/* Change	  */
       V1[i][u] = w;
       w  = X0[i][u] + hdt*(w + V0[i][u]);	/* New position component */
       mx = Math.max(mx, Math.abs(X1[i][u] - w));	/* Change	  */
       X1[i][u] = w;
     }
   return mx;
 }


/************************************************************************/
/*  Compute X1, V1 for one time step of deltat, using current Ac to	*/
/*  guess initial X1, V1 from current X0, V0.	X0, V0 are not altered.	*/
/*  Returns number of iterations required to converge (to tolerance	*/
/*  epsilon).  Stops in any case after `limit' iterations.		*/
/*  Increments `fail' on iterations which fail to converge.		*/

int oneStep()
 { double mx;  int ic;				/* Max change		*/
   mx = updateXV() + 2.0*epsilon;		/* Forces one iteration	*/
   for ( ic = 0;				/* Iteration counter	*/
	 (mx > epsilon) && (ic < limit);
	 ic++ )
    { greenAccel();  mx = updateXV(); }
   if ( mx > epsilon ) fail++;
   return ic;
 }


/************************************************************************/
/*  Carry out ns steps of integration using current parameters.		*/
/*  Ac should already have been calculated, perhaps from the previous	*/
/*  time step.  Leaves new sate in X0, V0, advances `stime'.		*/
/*  Returns total number of iterations, leaves `fail' with number of	*/
/*  failed iterations.							*/

public int nSteps(int ns)
 { int ic = 0;  double TX[][];  /* TX is used only for reference */
   fail = 0;
   for ( int k = 0;  k < ns;  k++ )
    { ic += oneStep();
      TX = X0;  X0 = X1;  X1 = TX;	/* "Swap" X0, X1  */
      TX = V0;  V0 = V1;  V1 = TX;	/*  and   V0, V1  */
    }
   stime += ns*deltat;
   return ic;
 }


/************************************************************************/
/*  Calculate Newtonian acceleration from current X0, leaving result	*/
/*  in Ac  Local variables are like those in accel() above.		*/

public void newtonAccel()
 { double Dij[], r, w, fi, fj;  int i, j, u;
   Dij = new double[3];  clearAc();
   /* Loop over pairs of particles */
   for ( i = 0;  i < Nc-1;  i++)
    for ( j = i+1;  j < Nc;  j++ )
     { r = 0.0;				/* To sum squares of displacements */
       for ( u = 0;  u < 3;  u++ )
        { w = X0[j][u] - X0[i][u];  r += w*w;  Dij[u] = w; }
       r = Math.sqrt(r);  w = r*r*r;
       fi = Mg[j]/w;   fj = Mg[i]/w;
       for ( u = 0;  u < 3;  u++ )
	{ Ac[i][u] += fi*Dij[u];  Ac[j][u] -= fj*Dij[u]; }
     }
 }


/************************************************************************/
/*  Initialize integration process with step dt, tolerance eps,		*/
/*  iteration limit lim.  May be also be used to change parameters	*/
/*  during integration.  Calculate Ac as Newtonian acceleration to	*/
/*  insure that Ac is available to begin subsequent integration.	*/

public void  initInteg(double dt, double eps, int lim)
 { deltat = dt;  epsilon = eps;  limit = lim;  newtonAccel(); }


/************************************************************************/
/*  Convert double to F10.6 format. */

static String f106(double d)
  { double ad = Math.abs(d);
    if ( (d >= 1000.0) || (d <= -100.0) ) return "***.******";
    else try
     { StringBuffer fbuff = new StringBuffer();  int i;
       long id = Math.round(d*1.0E6);
       fbuff.append(id);
       if ( (id >= 0) && (d < 100.0) ) fbuff.insert(0, ' ');
       if ( ad < 10.0 ) fbuff.insert(0, ' ');
       int l = fbuff.length();
       if ( l < 9 )
	 for ( i = 1;  i <= 9 - l;  i++ ) fbuff.insert(2, '0');
       fbuff.insert(3, '.');
       return fbuff.toString();
      }
     catch (/* StringIndexOutOfBounds */ Exception ex)
      { return "***.******"; }
  }


/************************************************************************/
/*  Convert int to I3 format */

static String i3(int i)
  { if ( i < -99 || i > 1000 ) return "***";
    else try
     { String istr = Integer.toString(i);
       int l = istr.length();
       StringBuffer ibuff = new StringBuffer(istr);
       for ( int k = 1;  k <= 3 - l;  k++ )  ibuff.insert(0, ' ');
       return ibuff.toString();
     }
    catch (/* StringIndexOutOfBounds */ Exception ex)
     { return "***"; }
  }


/************************************************************************/
/*  Print current state of integration using X0, V0.  Formats are	*/
/*  chosen for solar system examples.					*/

void printState(double X[][], double V[][] )
 { int i, j;
   System.out.println(
" i       X[x]       X[y]       X[z]       V[x]       V[y]       V[z]");
   for (i = 0;  i < Nc;  i++)
    { System.out.print(" " + i3(i));
      for ( j = 0;  j < 3;  j++ )
	System.out.print(" " + f106(X[i][j]) );
      for ( j = 0;  j < 3;  j++ )
	System.out.print(" " + f106(V[i][j]) );
      System.out.println();
    }
 }

/*  Line format:
 iii  -xx.xxxxxx -xx.xxxxxx -xx.xxxxxx  -vv.vvvvvx -vv.vvvvvx -vv.vvvvvv
*/

 }