/* Control panel for the rocket - the beginnings. 	*/
/*				S.F. Guo - April 1997	*/

import java.awt.*;

public class RControl extends Panel

 {
   static final double pi = 3.141592653589793;

   NBApplt nba;		/* The controlling applet		*/

   Rocket r;		/* The rocket				*/

   Status st;		/* Status 				*/

   Throttle th;		/* Throttle control			*/

   Attitude ats[];	/* Attitude control panels		*/
			/* See NBView.java for view parameters.	*/

   double cos[];        /* direction cosines			*/
   Object cosync;	/* To synchronize access to cos		*/


   /*  Initialize	*/

   public void init(NBApplt na, Rocket rk)
     { nba = na;  r = rk;
       setLayout(new GridLayout(5, 1, 5, 5));
       ats = new Attitude[2];
       cos = new double[3];  cosync = new Object();
       st  = new Status();  add(st);
       th  = new Throttle();  add(th);
       ats[0] = new Attitude();  add(ats[0]);
       ats[1] = new Attitude();  add(ats[1]);
       st.init(this);
       th.init(this);
       ats[0].init(this, 1, 2, 3); 
       ats[1].init(this, 1, 3, -2);
     }     


   /*  Set attitudes from current cosines.  Also sets rocket attitude.	*/
   /*  This should be the only procedure which sets the rocket attitude	*/
   /*  after this object is initialized.  To use it, synchronize on	*/
   /*  cosync, store the new cosines, and  setAtt().			*/

   public void setAtt()
     {
       for (int i = 0; i < ats.length; i++) ats[i].setAtt();
       synchronized ( r.TVsync )
	{ for (int u = 0; u < 3; u++) r.TV[u] = cos[u]; }
     }
        

   public void repaint()
     {
       paint(getGraphics ());
     }

   public void paint(Graphics gx)
     {
       ats[0].repaint();
       ats[1].repaint();
     }

 }


/*  Class for attitude indicator objects	*/

class Attitude extends Canvas 
 { 
   RControl   myctrl;

   static int    center =  70,
                 left   =   5,
                 width  = 130,
                 sz     = width + 2 * left;
       
   static double   rad  = (double)(width/2 - 2);
       
   static double   xc[] = {  0.0,  -20.0,   0.0,  20.0},
                   yc[] = {-60.0,   55.0,  35.0,  55.0};
   
   int    x[], y[];
   int    ax, ay, az, ix, iy, iz;       /* view specification  */


   /*  Initialize for view determined by vx, vy, vz.	*/

   void init(RControl myc, int vx, int vy, int vz)
     {
       /* setBackground(Color.white); */
       myctrl = myc;
       ax = vx; ay = vy; az = vz;
       ix = Math.abs(ax) -1; iy = Math.abs(ay) -1; iz = Math.abs(az) -1; 
       x = new int[4];
       y = new int[4];
       resize(sz, sz); 
     }


   /* Deal with relevant events */

   public boolean mouseDown(Event evt, int x, int y)
     { 
       double  xx  = x - center,
               yy  = y - center,
               lng = Math.sqrt(xx * xx + yy * yy);
       if (lng > rad)
          {
            xx  = xx * rad/lng; yy = yy * rad/lng;
            lng = rad;
          }
       
       if (ax < 0) xx = -xx; if (ay > 0) yy = -yy;
       xx = xx/rad; yy = yy/rad;
   synchronized (myctrl.cosync)
       {
       myctrl.cos[ix] = xx; myctrl.cos[iy] = yy; 
       double zs = 1.0;
       if (myctrl.cos[iz] < 0.0) zs = - 1.0;
       myctrl.cos[iz] = zs * Math.sqrt(Math.max(0.0, 1.0 - xx * xx -yy * yy));
       myctrl.setAtt();
       }
       return true;
     }
   
   public void setAtt ()
      {
       double xx =   myctrl.cos[ix]*rad, 
              yy = - myctrl.cos[iy]*rad;      
       if (ax < 0) xx = -xx; if (ay < 0) yy = -yy;
       setRot(xx, yy);
       repaint();

      }


   /*  Calculate scaling and rotation to screen coords x, y	*/

   public void setRot (double xx, double yy)   
      {
       /* System.out.println("setRot " + xx + "  " + yy); */

       double  lng = Math.sqrt(xx * xx + yy * yy);
       if (lng >= rad)
          {
            xx  = xx * rad/lng; yy = yy * rad/lng;
            lng = rad;
            rotate(xx/lng, -yy/lng, xc, yc);
          }
       else
        if (lng < rad)
          {
            double  xw[] = new double[4],
                    yw[] = new double[4];
            double  f    = lng/rad;
            
            for ( int i = 0; i < 4; i++)
               {
                 xw[i] = xc[i];
                 yw[i] = yc[i]*f;
               }
                
            rotate(xx/lng, -yy/lng, xw, yw);       
          }   
   }


   /*  Plane rotation by theta, where cos(theta) = c, sin(theta) = s	*/
   /*  Remember that screen coordinates are offset, the rotation is	*/
   /*  about screen center.						*/
      
   public void rotate(double s, double c, double xc[], double yc[])
     {
       for (int i = 0; i < 4; i++)
         { x[i] = (int) Math.round(c*xc[i] - s*yc[i]) + center;
           y[i] = (int) Math.round(s*xc[i] + c*yc[i]) + center;
         }
     }


   public void repaint()
     { paint(getGraphics ());
     }

   public void paint(Graphics gx)
     { gx.setColor(getBackground());
       gx.fillRect(0,0,sz,sz); 
       gx.setColor(Color.blue); 
       gx.drawOval(left, left, width, width);    
       gx.fillPolygon(x, y, 4);
      
     }

}


/*  Throttle control  */

class Throttle extends Canvas
 { 
   RControl myctrl;	/* My control	*/

   Rocket   r;		/* The rocket	*/

   double   throt,	/* Throttle setting  0 <= th <= 100%	*/
            prop;       /* Propellant        0 <= pr <= 100%    */

   Color tclr = Color.red,	/* For throttle setting */
         pclr = Color.green,    /* For propellant set   */
           fg = Color.black,	/* characters, box	*/
	 toff = Color.white;	/* Not throttle		*/
         
   int lb = 50, rb = 80,  tb = 19, bb = 121, ls = 10, ts = 30, bs = 121,
       lf = 90, rf = 120, xs = 132;  

   int ith = 0,		/* Screen position for throttle	  */
       ipr = 20;        /* Screen position for propellant */

   /*  Initialize the control.	*/

   void init(RControl rc)
    { myctrl = rc;  r = myctrl.r;
      throt = r.Thr;  ith = thri(throt);
      setProp();
      resize(140, 140);
    }


   /*  Calculate drawing height for throttle.	*/

   int thri(double th)
    { double thx = Math.min(100.0, Math.max(0.0, th));  /* in range */
      return ( bb - (int)Math.round(th) - 1 );
    }


    /*  Respond to mouse  */

    public boolean mouseDown(Event evt, int x, int y)
     { ith   = Math.min(bb-1, Math.max(tb+1, y));  /* in range */
       throt = bb - 1 - ith;
       synchronized ( r.thsync ) { r.Thr = throt;  r.throttle(); }
       repaint();
       return true;
     }

   /* adjust Propellant setting */

   public void setProp()
    { 
       int oldIpr = ipr; 
       prop  = Math.max(0.0, 100.0 * r.Pr/r.iPr);   ipr = thri(prop);
       if ( oldIpr != ipr)
	repaint();
    }

   /*  Paint the canvas	*/

   public void paint(Graphics gx)
    { gx.setColor(fg);
      gx.drawString("100%", ls, ts);
      gx.drawString("  0" , ls, bs);
      gx.drawString("Throt", lb-2, xs);
      gx.drawString("Prop" , lf, xs);
      gx.drawRect(lb, tb, rb-lb-1, bb-tb-1);
      gx.drawRect(lf, tb, rf-lf-1, bb-tb-1);
      gx.setColor(toff);
      gx.fillRect(lb+1, tb+1, rb-lb-2, ith-tb-1);
      gx.fillRect(lf+1, tb+1, rf-lf-2, ipr-tb-1);
      gx.setColor(tclr);
      gx.fillRect(lb+1, ith,  rb-lb-2, bb-ith-1);
      gx.setColor(pclr);
      gx.fillRect(lf+1, ipr,  rf-lf-2, bb-ipr-1);
    }

 }



/* Status panel */

class Status extends Panel
 { RControl myctrl;	/* Parent control 	*/
   Rocket   r;		/* Associated rocket	*/
   NBApplt  myap;	/* Controlling applet	*/

   double days;		/* Days in time unit	*/
   String tunit;	/* The time unit	*/

   String tstr = "",	/* Current time, units +	  */
	  dstr = "",	/* days.			  */
          astr = "";	/* Current perceived acceleration */

   /*  Left margin, vertical positions	*/

   int lm = 10,  ty = 20,  dy = 40,  ay = 60,  fy = 75,  sy = 105,
        rw = 60,  rh = 20,  sb = rh - 5,  sh = lm + 40, my = (sy + fy + rh)/2;
    
   Color fg = Color.black;

   Semaphore udtR;

   /* Initialize panel */

   void init(RControl myc)
    { myctrl = myc;  r = myc.r;  myap = myc.nba;  udtR = myap.udtR;
      days = NBSystems.days;
      tunit = " "+NBSystems.tunit+"s";
      resize(140, 140);
    }


   /* Set status for new time (tm in years, this converts to days ) */

   void setStatus(double tm)
    { double u = Math.floor(tm);
      tstr = u + tunit;
      dstr = Grnspan.f106((tm - u)*days) + " days" ;
      astr = Grnspan.f106(r.Accel) + " g";
      repaint();
    }

   public void paint(Graphics gx)
    { gx.setColor(fg);
      gx.drawString(tstr, lm, ty);
      gx.drawString(dstr, lm, dy);
      gx.drawString(astr, lm, ay);
      gx.drawRect(lm + 30, fy, rw, rh);
      gx.drawRect(lm + 30, sy, rw, rh);
      gx.drawString("Faster", sh, fy + sb);
      gx.drawString("Slower", sh, sy + sb); 
    }

   public boolean mouseDown(Event evt, int x, int y)
    { int du;
      if ( y < my ) du = 1; else du = -1;
      synchronized ( udtR ) { udtR.value += du; }
      repaint();
      return true;
    }



 }






