/*  Implements a group of 1-3 projections of a 3-D box, all as a common	*/
/*  scale.  Each view is implemented by an NBView.  The group is	*/
/*  responsible for scaling and coordinate transformation.  The views	*/
/*  map voxel coordinates to their projections.				*/
/*  									*/
/*  The proper sequence for setting up a group is:			*/
/*  									*/
/*	Contruct a ViewGroup vg						*/
/*	vg.init(...);							*/
/*	vg.addView(...);	Up to 3 views				*/
/*      vg.loadObject(...);	Up to NOb objects [view centers]	*/
/*      vg.setControls();       Set controls				*/
/*      vg.loadObject(...);	Up to NOb objects total [not centers]	*/
/*	vg.setView(...);	At beginning and to change view		*/
/*  									*/
/*  Initially the view center is fixed in space, typically at the	*/
/*  origin, but this can be changed to another point in space or	*/
/*  attached to one of the bodies.					*/

/*  See the end of this file for control menus.				*/


import java.awt.*;

public class ViewGroup extends Panel
 {
   public int   pw[],	/* Region dimensions (pixels) [x,y,z]	*/
		pc[],	/* Center pixels [x,y,z]		*/
	          nv,   /* Number of views allowed		*/
		  nc,	/* Number of views currently installed	*/
		 NOb,	/* Number of objects to view		*/
		 cob,	/* Number of objects currently viewed	*/
                cntr;	/* Index of central object, or -1	*/

   public double xwd,	/* Region width (space units) [x-axis]	*/
		 c[],	/* Region center			*/
	       scale,	/* space * scale = pixels		*/
              bscale;	/* body * bscale = pixels		*/

   Object vsafety;	/* Synchronization object		*/

   Semaphore prime;	/* Semaphore for signalling between	*/
			/* menus and views.  prime.value gives	*/
			/* number of mouse clicks remaining to	*/
			/* be recognized by views.  See PosMenu	*/
			/* below and NBView.			*/

   NBView vws[];	/* The views		*/

   NBObj  obs[];	/* The objects viewed	*/

   String where[] = {"Center", "South", "East"};

   ViewCtrl menus;	/* Choice menus for view control	*/

   Color fg = Color.black,	/* Standard foreground for labels	*/
         bg = Color.black;	/* Standard background for views	*/


   /*  Initialize for nvs views, volume dx by dy by dz voxels, 	*/
   /*  up to NOb plottable objects.				*/

   public void init(int nvs, int dx, int dy, int dz, int nob)
    { pw = new int[3];  pc = new int[3];  c = new double[3];
      vsafety = new Object();
      cntr = -1;	/* Initialized for coords fixed in space.	*/
      prime = new Semaphore();
      NOb = nob;  obs = new NBObj[NOb];
      nv = nvs;  vws = new NBView[nv];  nc = 0;  cob = 0;
      setLayout(new BorderLayout(5,5));
      pw[0] = dx;  pw[1] = dy;  pw[2] = dz;
      for (int i = 0;  i < 3;  i++)  pc[i] = pw[i]/2;
      menus = new ViewCtrl();  add("North", menus);
    }


   /*  Add view to group; vx, vy, vz determine view orientation and	*/
   /*  direction.							*/

   public void addView(int vx, int vy, int vz)
    { if ( nc < nv )
       { vws[nc] = new NBView(this, vx, vy, vz);  add(where[nc], vws[nc]);
         vws[nc].init();  nc++;
       }
    }


   /*  Load one object for viewing,  xx is array with position, */
   /*  bb is associated body.  Returns the resulting NBObj.	*/

   public NBObj loadObject(Body bb)
    { NBObj obj = new NBObj(bb);
      System.out.println("Loading "+bb.name);
      if ( cob < NOb )
       { obs[cob] = obj;  cob++;
	 for (int i = 0;  i < nc;  i++)
	   obj.ovw[i] = vws[i].loadObject(obj);
       }
      return obj;
    }


   /*  Set view controls, currently loaded objects will be available	*/
   /*  as view centers.							*/

   void setControls()
    { menus.init(this); }


   /*  Set (new) view parameters  (width, center-x, center-y, center-z)	*/
   /*  This doesn't do anything to the image.				*/

   public void setView(double vwd, double vcx, double vcy, double vcz)
    { synchronized (vsafety)
       { c[0] = vcx;  c[1] = vcy;  c[2] = vcz;
	 xwd = vwd;  scale = pw[0]/xwd;  cntr = -1;
	 bscale = scale/NBSystems.b2s;
	 for (int i = 0;  i < cob;  i++)	/* Rescale objects	*/
	  { Body b = obs[i].b;
	    obs[i].d0 = (int)Math.max(b.diam*bscale, b.icdiam);
	  }
       }
    }


   /*  Clear buffers */

   public void clear()
    { for (int i = 0;  i < nc;  i++ ) vws[i].clear();
    }


   /*  Load ob with space coords from xs[3] and compute voxel	*/
   /*  coordinates in vc[3].  To avoid integer overflow,	*/
   /*  voxel coordinates are limited to +/- 10^8 (100M).	*/

   static final double vxlimit = 1.0E8;

   void transform(double xs[], NBObj ob)
    { double xf;
      for (int i = 0;  i < 3; i++ )
       { ob.x[i] = xs[i];
	 xf = (xs[i] - c[i])*scale;
	      if (  xf > vxlimit ) xf =  vxlimit;
	 else if ( -xf > vxlimit ) xf = -vxlimit;
	 ob.c[i] = (int)Math.round(xf) + pc[i];
       }
    }


   /*  Set center coordinates from X[3].	*/

   void setCenter(double X[])
    { c[0] = X[0];  c[1] = X[1];  c[2] = X[2]; }


   /*  Clear then draw all objects.  Objects are brought up to date,	*/
   /*  using values for space coordinates in Xs.			*/

   public void drawAll(double Xs[][])
    { synchronized (vsafety)
       { clear();
         if (cntr >= 0) setCenter(Xs[cntr]);
	 for (int i = 0;  i < cob;  i++ )	/* Update objects	*/
	  { NBObj ob = obs[i];  transform(Xs[i], ob); }
	 for (int v = 0;  v < nc; v++ ) 	/* Update and draw views */
	      vws[v].updateDrawAll();
       }
    }


   /*  Recalculate voxel coordinates from current positions and	redraw	*/
   /*  buffers.  Used after view changes.  Note that view changes	*/
   /*  cannot affect depth ordering, though they can change the depths.	*/

   public void redrawAll()
    { if (cntr >=  0) setCenter(obs[cntr].x);
      for (int i = 0;  i < cob;  i++ )	/* Recalculate voxels	*/
       { NBObj ob = obs[i];
	 for (int k = 0;  k < 3;  k++)
	    ob.c[k] = (int)Math.round((ob.x[k] - c[k])*scale) + pc[k];
       }
      for (int v = 0;  v < nc; v++ ) 	/* Update and draw views */
	vws[v].updateDrawAll();
    }


   /*  Change view center to cx[3] (fixed in space) and redraw views.	*/

   public void recenter(double cx[])
    { synchronized (vsafety)
       { setCenter(cx);
	 cntr = -1;  clear();  redrawAll();
       }
      repaint();
    }


   /*  Change view center to voxel vx[3].  vx specifies displacement to	*/
   /*  new center relative to current center.				*/

   public void voxCenter(int vx[])
    { synchronized (vsafety)
       { System.out.println("vox: "+vx[0]+" "+vx[1]+" "+vx[2]);
         for (int i = 0;  i < 3;  i++)
          c[i] += vx[i]/scale;
	 cntr = -1;  clear();  redrawAll();
       }
      repaint();
    }


   /*  Change view center to obs[bi] and redraw views.	*/

   public void bodyCenter(int bi)
    { synchronized (vsafety)
       { cntr = bi; clear();  redrawAll(); }
      repaint();
    }


   /*  Change view width (x) to xw and redraw views.	*/

   public void rescale(double vwd)
    { synchronized (vsafety)
       { xwd = vwd;  scale = pw[0]/xwd;
	 bscale = scale/NBSystems.b2s;
	 for (int i = 0;  i < cob;  i++)	/* Rescale objects	*/
	  { Body b = obs[i].b;
	    obs[i].d0 = (int)Math.max(b.diam*bscale, b.icdiam);
	  }
	 clear();  redrawAll();
       }
      repaint();
    }


   /*  Repaint all views.	*/

   public void repaint()
    { synchronized (vsafety)
       { for (int i = 0;  i < nc;  i++) vws[i].repaint(); }
    }


 } /* End ViewGroup */



/************************************************************************/
/*  Choice menus to control the view group.  These interact fairly	*/
/*  intimately with ViewGroup.						*/

class ViewCtrl extends Panel
 { 
   ViewGroup myGroup;		/* Owner of this panel	*/

   PosMenu  position;		/* Choices for centering views	*/
   ScaleMenu vwscale;		/* Choices for scaling view	*/
   TextField message;		/* Instructions when needed	*/

   String dfltmsg = "Select choices to alter views.";


   /*  Initialize for ViewGroup myg.  All bodies defined in myg at	*/
   /*  initialization are taken as possible view centers, in addition	*/
   /*  to Origin and Space.						*/

   void init(ViewGroup myg)
    { myGroup = myg;
      position = new PosMenu();    add(position);  position.init(this);
      vwscale  = new ScaleMenu();  add(vwscale);   vwscale.init(this);
      for (int i = 0;  i < myGroup.cob;  i++)
	position.addItem(myGroup.obs[i].b.name);
      message  = new TextField(dfltmsg, 30);  add(message);
      message.setEditable(false);
    }


   /*  Install default message.	*/

   public void setDefault()
    { message.setText(dfltmsg); }

 }



/*  Center positions.  Choices are			*/
/*	origin						*/
/*	space (selected by mouse click),		*/
/*	first nc bodies of parent view (cf init)	*/

class PosMenu extends Choice
 {
   static double origin[] = {0.0, 0.0, 0.0};

   ViewCtrl   myCtrl;		/* Parent ViewCtrl		*/
   ViewGroup myGroup;		/* Grandparent group		*/
   Semaphore prime;		/* Semaphore to prime views for	*/
				/* mouse clicks			*/

   void init(ViewCtrl myc)
    { myCtrl = myc;  myGroup = myCtrl.myGroup;
      prime = myGroup.prime;
      addItem("Origin");
      addItem("Space");
    }



   /* Deal with relevant events */

   public boolean action(Event evt, Object arg)
    { Object tgt = evt.target;		/* The target object */
      if ( tgt == this )
       { int ind = getSelectedIndex();
	 switch ( ind )
	  { case 0:  myGroup.recenter(origin);  break;
	    case 1:  getSpaceCenter();  break;
           default:  myGroup.bodyCenter(ind - 2);
	  }
	 return true;
       }
      return false;
    }


    /* Select space center from mouse click in one of the views.	*/
    /* View first clicked will trigger recentering.			*/

    public void getSpaceCenter()
     { myCtrl.message.setText("Click in any view to select center.");
       synchronized (prime) { prime.value++; }
     }


 }


/*  Scale controls.  Choices are:	*/
/*	Much farther			*/
/*	Farther				*/
/*	Closer				*/
/*	Much closer			*/


class ScaleMenu extends Choice
 {
   ViewCtrl myCtrl;	/* Parent viewbar	*/

   void init(ViewCtrl myc)
    { myCtrl = myc;
      addItem("Much Farther");
      addItem("Farther");
      addItem("Closer");
      addItem("Much Closer");
    }

   /* Deal with relevant events */

   public boolean action(Event evt, Object arg)
    { Object tgt = evt.target;		/* The target object */
      if ( tgt == this )
       { double nsc;
	 switch ( getSelectedIndex() )
	  { case 0:  nsc = 8.000;  break;
	    case 1:  nsc = 2.000;  break;
	    case 2:  nsc = 0.500;  break;
	    case 3:  nsc = 0.125;  break;
           default:  System.out.println("Unexpected scale choice.");
		     nsc = 1.000;
	  }
         myCtrl.myGroup.rescale(nsc*myCtrl.myGroup.xwd);
	 return true;
       }
      return false;
    }

 }
